from pypika import Table

from .aggregate import Aggregate
from .substitution import Substitution
from .timeframe import TimeframePeriod


class AnalysisHelper:
    @staticmethod
    def db_table(name, **kwargs):
        try:
            schema, name = name.split(".")
            options = kwargs
            options["schema"] = schema
            return Table(name, **options)
        except ValueError:
            return Table(name, **kwargs)

    @staticmethod
    def columns_from_names(columns, table):
        return list(filter(lambda x: x["name"] in columns, table["columns"]))

    # TODO: dedup this non dry situation
    @staticmethod
    def aggregates_from_names(aggregates, table):
        return list(filter(lambda x: x["name"] in aggregates, table["aggregates"]))

    @classmethod
    def timeframe_periods_from_names(cls, timeframe_name, period_names, table):
        timeframe = next(
            timeframe
            for timeframe in table["timeframes"]
            if timeframe["name"] == timeframe_name
        )

        # filter out non-selected periods
        periods = [
            period for period in timeframe["periods"] if period["label"] in period_names
        ]

        return timeframe, periods

    @classmethod
    def columns(cls, columns, db_table):
        return [cls.field_from_column(d, db_table) for d in columns]

    @classmethod
    def aggregates(cls, aggregates, db_table):
        return [
            cls.field_from_aggregate(aggregate, db_table) for aggregate in aggregates
        ]

    @classmethod
    def periods(cls, timeframes, db_table):
        return [
            cls.field_from_timeframe_period(timeframe["timeframe"], period, db_table)
            for timeframe in timeframes
            for period in timeframe["periods"]
        ]

    @classmethod
    def field_from_aggregate(cls, aggregate, db_table):
        return Aggregate(aggregate, db_table).sql

    @classmethod
    def field_from_column(cls, d, db_table):
        return Substitution(d["sql"], db_table, column=d).sql

    @classmethod
    def field_from_timeframe_period(cls, timeframe, period, db_table):
        return TimeframePeriod(timeframe, period, db_table).sql



import importlib
from typing import Union

from . import BasePlugin, PluginDefinition, PluginType, Variant


def lazy_import(module, cls):
    def lazy():
        return getattr(importlib.import_module(module, "meltano.core.plugin"), cls)

    return lazy


base_plugin_classes = {
    PluginType.EXTRACTORS: lazy_import(".singer", "SingerTap"),
    PluginType.LOADERS: lazy_import(".singer", "SingerTarget"),
    PluginType.TRANSFORMS: lazy_import(".dbt", "DbtTransformPlugin"),
    PluginType.MODELS: lazy_import(".model", "ModelPlugin"),
    PluginType.DASHBOARDS: lazy_import(".dashboard", "DashboardPlugin"),
    PluginType.ORCHESTRATORS: lazy_import(".airflow", "Airflow"),
    PluginType.TRANSFORMERS: lazy_import(".dbt", "DbtPlugin"),
    PluginType.FILES: lazy_import(".file", "FilePlugin"),
}


def base_plugin_factory(
    plugin_def: PluginDefinition, variant_or_name: Union[str, Variant]
) -> BasePlugin:
    plugin_cls = base_plugin_classes.get(plugin_def.type, lambda: BasePlugin)()
    variant = plugin_def.find_variant(variant_or_name)
    return plugin_cls(plugin_def, variant)



import logging
from collections import namedtuple
from datetime import date, datetime
from typing import Optional

from .db import project_engine
from .meltano_invoker import MeltanoInvoker
from .plugin import PluginRef, PluginType
from .plugin.settings_service import PluginSettingsService, SettingMissingError
from .plugin_discovery_service import PluginDiscoveryService, PluginNotFoundError
from .project import Project
from .project_plugins_service import ProjectPluginsService
from .schedule import Schedule
from .utils import NotFound, coerce_datetime, find_named, iso8601_datetime, nest


class ScheduleAlreadyExistsError(Exception):
    """Occurs when a schedule already exists."""

    def __init__(self, schedule):
        self.schedule = schedule


class ScheduleDoesNotExistError(Exception):
    """Occurs when a schedule does not exist."""

    def __init__(self, name):
        self.name = name


class ScheduleNotFoundError(Exception):
    """Occurs when a schedule for a namespace cannot be found."""

    def __init__(self, namespace):
        self.namespace = namespace


class ScheduleService:
    def __init__(self, project: Project, plugins_service: ProjectPluginsService = None):
        self.project = project
        self.plugins_service = plugins_service or ProjectPluginsService(project)

    def add(
        self,
        session,
        name,
        extractor: str,
        loader: str,
        transform: str,
        interval: str,
        start_date: Optional[datetime] = None,
        **env,
    ):
        start_date = coerce_datetime(start_date) or self.default_start_date(  # TODO
            session, extractor
        )

        schedule = Schedule(
            name, extractor, loader, transform, interval, start_date, env=env
        )

        return self.add_schedule(schedule)

    def remove(self, name):
        return self.remove_schedule(name)

    def default_start_date(self, session, extractor: str) -> datetime:
        """
        Returns the `start_date` of the extractor, or now.
        """
        extractor = self.plugins_service.find_plugin(
            extractor, plugin_type=PluginType.EXTRACTORS
        )
        start_date = None
        try:
            settings_service = PluginSettingsService(
                self.project, extractor, plugins_service=self.plugins_service
            )
            start_date = settings_service.get("start_date", session=session)
        except SettingMissingError:
            logging.debug(f"`start_date` not found in {extractor}.")

        # TODO: this coercion should be handled by the `kind` attribute
        # on the actual setting
        if isinstance(start_date, date):
            return coerce_datetime(start_date)

        if isinstance(start_date, datetime):
            return start_date

        return iso8601_datetime(start_date) or datetime.utcnow()

    def add_schedule(self, schedule: Schedule):
        with self.project.meltano_update() as meltano:
            # guard if it already exists
            if schedule in meltano.schedules:
                raise ScheduleAlreadyExistsError(schedule)

            meltano.schedules.append(schedule)

        return schedule

    def remove_schedule(self, name: str):
        with self.project.meltano_update() as meltano:
            try:
                # guard if it doesn't exist
                schedule = find_named(self.schedules(), name)
            except NotFound:
                raise ScheduleDoesNotExistError(name)

            # find the schedules plugin config
            meltano.schedules.remove(schedule)

        return name

    def update_schedule(self, schedule: Schedule):
        with self.project.meltano_update() as meltano:
            try:
                idx = meltano.schedules.index(schedule)
                meltano.schedules[idx] = schedule
            except ValueError:
                raise ScheduleDoesNotExistError(schedule.name)

    def find_namespace_schedule(self, namespace: str) -> Schedule:

        """
        Search for a Schedule that runs for a certain plugin namespace.
        For instance, `tap_carbon` would yield the first schedule that
        runs for the `tap-carbon` extractor.
        """

        try:
            extractor = (
                self.plugins_service.discovery_service.find_definition_by_namespace(
                    PluginType.EXTRACTORS, namespace
                )
            )

            return next(
                schedule
                for schedule in self.schedules()
                if schedule.extractor == extractor.name
            )
        except (PluginNotFoundError, StopIteration) as err:
            raise ScheduleNotFoundError(namespace) from err

    def schedules(self):
        return self.project.meltano.schedules

    def find_schedule(self, name):
        try:
            return find_named(self.schedules(), name)
        except StopIteration as err:
            raise ScheduleNotFoundError(name) from err

    def run(self, schedule, *args, env={}, **kwargs):
        return MeltanoInvoker(self.project).invoke(
            ["elt", *schedule.elt_args, *args], env={**schedule.env, **env}, **kwargs
        )



import logging
import subprocess
from functools import wraps
from urllib.parse import urlsplit

import meltano
import requests
from flask import (
    Blueprint,
    current_app,
    g,
    jsonify,
    make_response,
    redirect,
    render_template,
    request,
)
from flask_login import current_user
from flask_security import logout_user, roles_required
from jinja2 import TemplateNotFound
from meltano.api.api_blueprint import APIBlueprint
from meltano.api.security.auth import block_if_readonly, passes_authentication_checks
from meltano.core.project import Project, ProjectReadonly
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.utils import truthy

logger = logging.getLogger(__name__)
root = Blueprint("root", __name__)


def redirect_to_login_if_auth_required(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if passes_authentication_checks():
            return f(*args, **kwargs)

        return current_app.login_manager.unauthorized()

    return decorated


@root.route("/-/embed/", defaults={"token": ""})
@root.route("/-/embed/<token>")
def embed(token):
    try:
        return render_template("embed.html", jsContext=g.jsContext)
    except TemplateNotFound:
        return "Please run `make bundle` from src/webapp of the Meltano project."


# this route is a catch-all route to forward
# all oustanding request (not caught by any route)
# to the front-end.
@root.route("/", defaults={"path": ""})
@root.route("/<path:path>")
@redirect_to_login_if_auth_required
def default(path):
    try:
        return render_template("webapp.html", jsContext=g.jsContext)
    except TemplateNotFound:
        return "Please run `make bundle` from src/webapp of the Meltano project."


@root.route("/bootstrap")
@redirect_to_login_if_auth_required
def bootstrap():
    return redirect(current_app.config["MELTANO_UI_URL"])


@root.route("/echo", methods=["POST"])
def echo():
    payload = request.get_json()
    print(payload)
    return jsonify(payload)


api_root = APIBlueprint("api_root", __name__, url_prefix="/api/v1/")


@api_root.route("/version")
def version():
    response_payload = {"version": meltano.__version__}

    if truthy(request.args.get("include_latest")):
        res = requests.get("https://pypi.org/pypi/meltano/json")
        pypi_payload = res.json()
        response_payload["latest_version"] = pypi_payload["info"]["version"]

    return jsonify(response_payload)


@api_root.route("/upgrade", methods=["POST"])
@roles_required("admin")
@block_if_readonly
def upgrade():
    meltano.api.executor.upgrade()
    return "Meltano update in progress.", 201


@api_root.route("/identity")
def identity():
    project = Project.find()
    settings_service = ProjectSettingsService(project)

    if current_user.is_anonymous:
        return jsonify(
            {
                "username": "Anonymous",
                "anonymous": True,
                "can_sign_in": settings_service.get("ui.authentication"),
            }
        )

    return jsonify(
        {"username": current_user.username, "anonymous": False, "can_sign_in": False}
    )


@api_root.route("/health")
def health():
    return jsonify({"healthy": True})



import asyncio
import json
import logging
import os
import shutil
import subprocess
import sys
from datetime import datetime
from enum import IntFlag
from pathlib import Path

from meltano.core.elt_context import ELTContext
from meltano.core.job import Job, Payload
from meltano.core.logging import capture_subprocess_output
from meltano.core.plugin import PluginType
from meltano.core.plugin.singer import PluginType, SingerTap, SingerTarget
from meltano.core.plugin_invoker import PluginInvoker, invoker_factory
from meltano.core.utils import file_has_data

from . import Runner, RunnerError


class BookmarkWriter:
    def __init__(self, job, session, payload_flag=Payload.STATE):
        self.job = job
        self.session = session
        self.payload_flag = payload_flag

    def writeline(self, line):
        if self.job is None:
            logging.info(
                f"Running outside a Job context: incremental state could not be updated."
            )
            return

        try:
            new_state = json.loads(line)
            job = self.job

            job.payload["singer_state"] = new_state
            job.payload_flags |= self.payload_flag
            job.save(self.session)

            logging.info(f"Incremental state has been updated at {datetime.utcnow()}.")
            logging.debug(f"Incremental state: {new_state}")
        except Exception as err:
            logging.warning(
                "Received state is invalid, incremental state has not been updated"
            )

    def flush(self):
        pass


class SingerRunner(Runner):
    def __init__(self, elt_context: ELTContext):
        self.context = elt_context

    def stop(self, process, **wait_args):
        while True:
            try:
                code = process.wait(**wait_args)
                logging.debug(f"{process} exited with {code}")
                return code
            except subprocess.TimeoutExpired:
                process.kill()
                logging.error(f"{process} was killed.")

    async def invoke(  # noqa: WPS217, WPS210, WPS211, WPS213, WPS231
        self,
        tap: PluginInvoker,
        target: PluginInvoker,
        extractor_log=None,
        loader_log=None,
        extractor_out=None,
        loader_out=None,
    ):
        """Invoke tap and target together."""
        extractor_log = extractor_log or sys.stderr
        loader_log = loader_log or sys.stderr

        # Start tap
        try:
            p_tap = await tap.invoke_async(
                stdout=asyncio.subprocess.PIPE,  # Singer messages
                stderr=asyncio.subprocess.PIPE,  # Log
            )
        except Exception as err:
            raise RunnerError(f"Cannot start extractor: {err}") from err

        # Start target
        try:
            p_target = await target.invoke_async(
                stdin=asyncio.subprocess.PIPE,  # Singer messages
                stdout=asyncio.subprocess.PIPE,  # Singer state
                stderr=asyncio.subprocess.PIPE,  # Log
            )
        except Exception as err:
            raise RunnerError(f"Cannot start loader: {err}") from err

        # Process tap output
        tap_outputs = [p_target.stdin]
        if extractor_out:
            tap_outputs.insert(0, extractor_out)

        tap_stdout_future = asyncio.ensure_future(
            capture_subprocess_output(p_tap.stdout, *tap_outputs)
        )
        tap_stderr_future = asyncio.ensure_future(
            capture_subprocess_output(p_tap.stderr, extractor_log)
        )

        # Process target output
        target_outputs = [self.bookmark_writer()]
        if loader_out:
            target_outputs.insert(0, loader_out)

        target_stdout_future = asyncio.ensure_future(
            capture_subprocess_output(p_target.stdout, *target_outputs)
        )
        target_stderr_future = asyncio.ensure_future(
            capture_subprocess_output(p_target.stderr, loader_log)
        )

        # Wait for tap or target to complete
        tap_process_future = asyncio.ensure_future(p_tap.wait())
        target_process_future = asyncio.ensure_future(p_target.wait())
        done, _ = await asyncio.wait(
            [tap_process_future, target_process_future],
            return_when=asyncio.FIRST_COMPLETED,
        )

        if target_process_future in done:
            # If the target completes before the tap, it failed before processing all tap output
            target_code = target_process_future.result()

            # Kill tap and cancel output processing since there's no more target to forward messages to
            p_tap.kill()
            tap_stdout_future.cancel()
            tap_stderr_future.cancel()

            # Pretend the tap finished successfully since it didn't itself fail
            tap_code = 0

            # Wait for all buffered target output to be processed
            await asyncio.wait([target_stdout_future, target_stderr_future])
        else:  # if tap_process_future in done:
            # If the tap completes before the target, the target should have a chance to process all tap output
            tap_code = tap_process_future.result()

            # Wait for all buffered tap output to be processed
            await asyncio.wait([tap_stdout_future, tap_stderr_future])

            # Close target stdin so process can complete naturally
            p_target.stdin.close()

            # Wait for all buffered target output to be processed
            await asyncio.wait([target_stdout_future, target_stderr_future])

            # Wait for target to complete
            target_code = await target_process_future

        if tap_code and target_code:
            raise RunnerError(
                "Extractor and loader failed",
                {PluginType.EXTRACTORS: tap_code, PluginType.LOADERS: target_code},
            )
        elif tap_code:
            raise RunnerError("Extractor failed", {PluginType.EXTRACTORS: tap_code})
        elif target_code:
            raise RunnerError("Loader failed", {PluginType.LOADERS: target_code})

    def bookmark_writer(self):
        incomplete_state = self.context.full_refresh and self.context.select_filter
        payload_flag = Payload.INCOMPLETE_STATE if incomplete_state else Payload.STATE
        return BookmarkWriter(
            self.context.job, self.context.session, payload_flag=payload_flag
        )

    def dry_run(self, tap: PluginInvoker, target: PluginInvoker):
        logging.info("Dry run:")
        logging.info(f"\textractor: {tap.plugin.name} at '{tap.exec_path()}'")
        logging.info(f"\tloader: {target.plugin.name} at '{target.exec_path()}'")

    async def run(
        self, extractor_log=None, loader_log=None, extractor_out=None, loader_out=None
    ):
        tap = self.context.extractor_invoker()
        target = self.context.loader_invoker()

        if self.context.dry_run:
            return self.dry_run(tap, target)

        with tap.prepared(self.context.session), target.prepared(self.context.session):
            await self.invoke(
                tap,
                target,
                extractor_log=extractor_log,
                loader_log=loader_log,
                extractor_out=extractor_out,
                loader_out=loader_out,
            )



import logging
from datetime import datetime
from fnmatch import fnmatch
from functools import wraps

from flask import current_app, jsonify, request
from flask_login import current_user
from flask_principal import Need, Permission
from flask_security import auth_required
from meltano.api.models import db
from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from werkzeug.exceptions import Forbidden

from .identity import FreeUser

HTTP_READONLY_CODE = 499


class ResourcePermission(Permission):
    def in_scope(self, need, scopes):
        matches = (fnmatch(need.value, scope.value) for scope in scopes)

        return any(matches)

    def allows(self, identity):
        logging.debug(f"Authorizing {self} using {identity}")
        needs_scoped = []
        for need in self.needs:
            # included
            scopes = (
                scope
                for scope in identity.provides
                if fnmatch(need.method, scope.method)
            )
            logging.debug(f"Found include scope {scopes} in {identity}")

            scoped = self.in_scope(need, scopes)
            needs_scoped.append(scoped)

        for excluded in self.excludes:
            # excluded
            scopes = (
                scope
                for scope in identity.provides
                if fnmatch(excluded.method, scope.method)
            )
            logging.debug(f"Found exclude scope {scopes} in {identity}")

            if self.in_scope(need, scopes):
                return False

        if all(needs_scoped):
            return True

        return False


def permission_for(permission_type, context):
    return ResourcePermission(Need(permission_type, context))


def permit(permission_type, context):
    if not permission_for(permission_type, context).can():
        raise Forbidden()


def passes_authentication_checks():
    project = Project.find()
    settings_service = ProjectSettingsService(project)

    if not settings_service.get("ui.authentication"):
        logging.debug(f"Authentication not required because it's disabled")
        return True

    if current_user.is_authenticated:
        logging.debug(f"Authenticated as '{current_user.username}'")
        return True

    if settings_service.get("ui.anonymous_readonly") and current_user.is_anonymous:
        # The `@roles_required("admin")` and `@block_if_readonly` checks
        # will take care of enforcing authentication as appropriate
        logging.debug(
            f"Authentication not required because anonymous users have read-only access"
        )
        return True

    return False


def block_if_api_auth_required(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if request.method == "OPTIONS" or passes_authentication_checks():
            return f(*args, **kwargs)

        return "Authentication is required to access this resource.", 401

    return decorated


def block_if_readonly(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        project = Project.find()
        settings_service = ProjectSettingsService(project)

        if settings_service.get("ui.readonly"):
            return (
                jsonify(
                    {"error": True, "code": "Meltano UI is running in read-only mode"}
                ),
                HTTP_READONLY_CODE,
            )

        if settings_service.get("ui.anonymous_readonly") and current_user.is_anonymous:
            return (
                jsonify(
                    {
                        "error": True,
                        "code": "Meltano UI is running in read-only mode until you sign in",
                    }
                ),
                HTTP_READONLY_CODE,
            )

        return f(*args, **kwargs)

    return decorated


def unauthorized_callback():
    """
    Meltano is mainly an API, so let's return plain 403 (Forbidden)
    instead of redirecting anywhere.
    """

    return "You do not have the required permissions.", 403


def _identity_loaded_hook(sender, identity):
    """
    Meltano uses a resource permission scheme.
    This hook will add the specific Permission
    to the current identity.
    """

    # something weird is going on here when testing
    # SQLAlchemy complains about the roles not being
    # in a session.
    for role in current_user.roles:
        db.session.add(role)

    # each permission is a Need(permission_type, context),
    # i.e. ("view", "finance.*", "design")
    for perm in (perm for role in current_user.roles for perm in role.permissions):
        perm = Need(perm.type, perm.context)

        identity.provides.add(perm)


def _user_logged_in_hook(sender, user):
    """
    Update the audit columns for the User
    """
    user.last_login_at = datetime.utcnow()
    user.last_activity_at = datetime.utcnow()
    user.login_count += 1



import errno
import logging
import os
import sys
import threading
from contextlib import contextmanager
from copy import deepcopy
from functools import wraps
from pathlib import Path
from typing import Dict, Union

import fasteners
import yaml
from atomicwrites import atomic_write
from dotenv import dotenv_values
from werkzeug.utils import secure_filename

from .behavior.versioned import Versioned
from .error import Error
from .meltano_file import MeltanoFile
from .utils import makedirs, truthy

logger = logging.getLogger(__name__)


PROJECT_ROOT_ENV = "MELTANO_PROJECT_ROOT"
PROJECT_READONLY_ENV = "MELTANO_PROJECT_READONLY"


class ProjectNotFound(Error):
    def __init__(self, project):
        super().__init__(
            f"Cannot find `{project.meltanofile}`. Are you in a meltano project?"
        )


class ProjectReadonly(Error):
    def __init__(self):
        super().__init__(f"This Meltano project is deployed as read-only")


class Project(Versioned):
    """
    Represent the current Meltano project from a file-system
    perspective.
    """

    __version__ = 1
    _activate_lock = threading.Lock()
    _find_lock = threading.Lock()
    _meltano_rw_lock = fasteners.ReaderWriterLock()
    _default = None

    def __init__(self, root: Union[Path, str]):
        self.root = Path(root).resolve()

        self.readonly = False

        self.__meltano_ip_lock = None

    @property
    def _meltano_ip_lock(self):
        if self.__meltano_ip_lock is None:
            self.__meltano_ip_lock = fasteners.InterProcessLock(
                self.run_dir("meltano.yml.lock")
            )

        return self.__meltano_ip_lock

    @property
    def env(self):
        return {PROJECT_ROOT_ENV: str(self.root)}

    @classmethod
    @fasteners.locked(lock="_activate_lock")
    def activate(cls, project: "Project"):
        project.ensure_compatible()

        # create a symlink to our current binary
        try:
            executable = Path(os.path.dirname(sys.executable), "meltano")
            if executable.is_file():
                project.run_dir().joinpath("bin").symlink_to(executable)
        except FileExistsError:
            pass
        except OSError as e:
            if e.errno == errno.EOPNOTSUPP:
                logger.warning(
                    f"Could not create symlink: {e}\nPlease make sure that the underlying filesystem supports symlinks."
                )
            else:
                raise

        logger.debug(f"Activated project at {project.root}")

        # set the default project
        cls._default = project

    @classmethod
    def deactivate(cls):
        cls._default = None

    @property
    def file_version(self):
        return self.meltano.version

    @classmethod
    @fasteners.locked(lock="_find_lock")
    def find(cls, project_root: Union[Path, str] = None, activate=True):
        if cls._default:
            return cls._default

        project_root = project_root or os.getenv(PROJECT_ROOT_ENV) or os.getcwd()
        project = Project(project_root)

        if not project.meltanofile.exists():
            raise ProjectNotFound(project)

        # if we activate a project using `find()`, it should
        # be set as the default project for future `find()`
        if activate:
            cls.activate(project)

        if truthy(os.getenv(PROJECT_READONLY_ENV, "false")):
            project.readonly = True

        return project

    @property
    def meltano(self) -> Dict:
        """Return a copy of the current meltano config"""
        # fmt: off
        with self._meltano_rw_lock.read_lock(), \
             self.meltanofile.open() as meltanofile:
            return MeltanoFile.parse(yaml.safe_load(meltanofile))
        # fmt: on

    @contextmanager
    def meltano_update(self):
        """
        Yield the current meltano configuration and update the meltanofile
        if the context ends gracefully.
        """

        if self.readonly:
            raise ProjectReadonly

        # fmt: off
        with self._meltano_rw_lock.write_lock(), \
            self._meltano_ip_lock:

            with self.meltanofile.open() as meltanofile:
                # read the latest version
                meltano_update = MeltanoFile.parse(yaml.safe_load(meltanofile))

            yield meltano_update

            try:
                with atomic_write(self.meltanofile, overwrite=True) as meltanofile:
                    # update if everything is fine
                    yaml.dump(meltano_update, meltanofile, default_flow_style=False, sort_keys=False)
            except Exception as err:
                logger.critical(f"Could not update meltano.yml: {err}")
                raise
        # fmt: on

    def root_dir(self, *joinpaths):
        return self.root.joinpath(*joinpaths)

    @contextmanager
    def file_update(self):
        if self.readonly:
            raise ProjectReadonly

        yield self.root

    @property
    def meltanofile(self):
        return self.root.joinpath("meltano.yml")

    @property
    def dotenv(self):
        return self.root.joinpath(".env")

    @property
    def dotenv_env(self):
        return dotenv_values(self.dotenv)

    @contextmanager
    def dotenv_update(self):
        if self.readonly:
            raise ProjectReadonly

        yield self.dotenv

    @makedirs
    def meltano_dir(self, *joinpaths):
        return self.root.joinpath(".meltano", *joinpaths)

    @makedirs
    def analyze_dir(self, *joinpaths):
        return self.root_dir("analyze", *joinpaths)

    @makedirs
    def extract_dir(self, *joinpaths):
        return self.root_dir("extract", *joinpaths)

    @makedirs
    def venvs_dir(self, *prefixes):
        return self.meltano_dir(*prefixes, "venv")

    @makedirs
    def run_dir(self, *joinpaths):
        return self.meltano_dir("run", *joinpaths)

    @makedirs
    def logs_dir(self, *joinpaths):
        return self.meltano_dir("logs", *joinpaths)

    @makedirs
    def job_dir(self, job_id, *joinpaths):
        return self.run_dir("elt", secure_filename(job_id), *joinpaths)

    @makedirs
    def job_logs_dir(self, job_id, *joinpaths):
        return self.logs_dir("elt", secure_filename(job_id), *joinpaths)

    @makedirs
    def model_dir(self, *joinpaths):
        return self.meltano_dir("models", *joinpaths)

    @makedirs
    def plugin_dir(self, plugin: "PluginRef", *joinpaths):
        return self.meltano_dir(plugin.type, plugin.name, *joinpaths)

    def __eq__(self, other):
        return self.root == other.root

    def __hash__(self):
        return self.root.__hash__()



import logging
from collections import namedtuple
from copy import deepcopy
from enum import Enum
from functools import singledispatch
from typing import Dict

import networkx as nx
import sqlparse
from meltano.core.behavior.visitor import visit_with
from networkx.readwrite import json_graph
from sqlparse.sql import Comparison, TokenList

from .analysis_helper import AnalysisHelper
from .base import MeltanoDesign, MeltanoTable


class InvalidIdentifier(Exception):
    pass


Identifier = namedtuple("Identifier", ("schema", "table", "field", "alias"))

DimensionGroup = namedtuple("DimensionGroup", ("dimensions", "group"))

ComparisonFields = namedtuple("ComparisonFields", ("left", "right"))


# Visitor pattern on the SQL AST
@singledispatch
def visit(node, executor, depth=0):
    logging.debug(f"Visiting node {node}")


@visit.register(TokenList)
def _(node: TokenList, executor, depth=0):
    logging.debug(f"Visiting node {node}")
    for t in node.tokens:
        executor.visit(t, depth + 1)


@visit.register(Comparison)
def _(node: Comparison, executor, depth=0):
    logging.debug(f"Visiting node {node}")
    executor.comparison(node, depth)


@visit_with(visit)
class PypikaJoinExecutor:
    def __init__(self, design, join):
        self.design = MeltanoDesign(design)
        self.join = join
        self.result = None
        self.comparison_fields = None

    def set_comparison_fields(self, left, right):
        self.comparison_fields = ComparisonFields(left, right)

    def comparison(self, node, depth):
        table = self.join["related_table"]

        left = self.parse_identifier(node.left)
        right = self.parse_identifier(node.right)

        left_table = self.design.find_table(left.table)
        right_table = self.design.find_table(right.table)

        left_field = getattr(
            AnalysisHelper.db_table(
                left_table.sql_table_name, alias=left_table.find_source_name()
            ),
            left.field,
        )

        right_field = getattr(
            AnalysisHelper.db_table(
                right_table.sql_table_name, alias=right_table.find_source_name()
            ),
            right.field,
        )

        self.set_comparison_fields(left, right)

        # TODO: do a stack based approach to handle complex cases
        self.result = left_field == right_field

    @classmethod
    def parse_identifier(cls, identifier) -> Identifier:
        token_values = map(lambda t: t.value, identifier.tokens)

        # name, field
        if len(identifier.tokens) == 3:
            table, _, field = token_values
            return Identifier(None, table, field, None)

        # schema, name, field
        if len(identifier.tokens) == 5:
            schema, _, table, _, field = token_values
            return Identifier(schema, table, field, None)

        # schema, name, field, alias
        if len(identifier.tokens) == 7:
            schema, _, table, _, field, _, alias = token_values
            return Identifier(schema, table, field, alias)

        raise InvalidIdentifier(str(identifier))


class JoinNotFound(Exception):
    pass


class JoinType(Enum):
    left_join = "left"
    inner_join = "inner"
    full_outer_join = "full_outer"
    cross_join = "cross"


class DesignHelper:
    def __init__(self, design):
        self.design = design

    @property
    def name(self):
        return self.design["name"]

    @property
    def joins(self):
        return deepcopy(self.design["joins"])

    @property
    def tables(self):
        tables = [self.design["related_table"]]
        tables += [j["related_table"] for j in self.design["joins"]]
        return list(map(MeltanoTable, tables))

    @property
    def base_table_name(self):
        return self.design["related_table"]["name"]

    def joins_for_table(self, table_name: str):
        # get the graph for networkx
        G = json_graph.node_link_graph(self.design["graph"])
        return nx.shortest_path(G, source=self.base_table_name, target=table_name)

    def join_for(self, join_selection: Dict):
        join = self.get_join(self, join_selection["name"])
        table = join["related_table"]

        db_table = AnalysisHelper.db_table(table["sql_table_name"], alias=join["name"])
        selected = {"columns": [], "aggregates": [], "timeframes": []}

        try:
            selected["columns"] = AnalysisHelper.columns_from_names(
                join_selection["columns"], table
            )
        except KeyError:
            pass

        try:
            selected["aggregates"] = AnalysisHelper.aggregates_from_names(
                join_selection["aggregates"], table
            )
        except KeyError:
            pass

        try:
            selected["timeframes"] = [
                self.timeframe_periods_for(table, timeframe_selection)
                for timeframe_selection in join_selection["timeframes"]
            ]
        except KeyError:
            pass

        join_executor = PypikaJoinExecutor(self.design, join)
        join_executor.visit(sqlparse.parse(join["sql_on"])[0])

        return {
            "table": table,
            "db_table": db_table,
            "on": join_executor.result,
            "join": join,
            "name": join_selection["name"],
            **selected,
        }

    def timeframe_periods_for(self, table, timeframe_selection):
        timeframe_name = timeframe_selection["name"]
        period_names = [p["label"] for p in timeframe_selection["periods"]]

        db_table = AnalysisHelper.db_table(
            table["sql_table_name"], alias=timeframe_name
        )
        timeframe, periods = AnalysisHelper.timeframe_periods_from_names(
            timeframe_name, period_names, table
        )

        return {
            "table": table,
            "db_table": db_table,
            "timeframe": timeframe,
            "periods": periods,
            "period_labels": period_names,
        }

    @classmethod
    def get_join(cls, design, name):
        try:
            return next(join for join in design.joins if join["name"] == name)
        except StopIteration:
            raise JoinNotFound(
                f"No join named '{name}' was found in design '{design.name}'."
            )

    def __getitem__(self, idx):
        return self.design[idx]



import configparser
import logging
import os
import subprocess

from meltano.core.behavior.hookable import hook
from meltano.core.error import SubprocessError
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.utils import nest

from . import BasePlugin, PluginType


class AirflowInvoker(PluginInvoker):
    def env(self):
        env = super().env()

        env["AIRFLOW_HOME"] = str(self.plugin_config_service.run_dir)
        env["AIRFLOW_CONFIG"] = str(self.files["config"])

        return env


class Airflow(BasePlugin):
    __plugin_type__ = PluginType.ORCHESTRATORS

    invoker_class = AirflowInvoker

    @property
    def config_files(self):
        return {"config": "airflow.cfg"}

    def process_config(self, flat_config):
        config = {}
        for key, value in flat_config.items():
            nest(config, key, str(value))
        return config

    @hook("before_install")
    def setup_env(self, *args, **kwargs):
        # to make airflow installables without GPL dependency
        os.environ["SLUGIFY_USES_TEXT_UNIDECODE"] = "yes"

    @hook("before_configure")
    def before_configure(self, invoker, session):
        # generate the default `airflow.cfg`
        handle = invoker.invoke(
            "--help",
            require_preparation=False,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        handle.wait()

        airflow_cfg_path = invoker.files["config"]
        logging.debug(f"Generated default '{str(airflow_cfg_path)}'")

        # open the configuration and update it
        # now we let's update the config to use our stubs
        airflow_cfg = configparser.ConfigParser()

        with airflow_cfg_path.open() as cfg:
            airflow_cfg.read_file(cfg)
            logging.debug(f"Loaded '{str(airflow_cfg_path)}'")

        config = invoker.plugin_config_processed
        for section, cfg in config.items():
            airflow_cfg[section].update(cfg)
            logging.debug(f"\tUpdated section [{section}] with {cfg}")

        with airflow_cfg_path.open("w") as cfg:
            airflow_cfg.write(cfg)
            logging.debug(f"Saved '{str(airflow_cfg_path)}'")

        # we've changed the configuration here, so we need to call
        # prepare again on the invoker so it re-reads the configuration
        # for the Airflow plugin
        invoker.prepare(session)
        handle = invoker.invoke(
            "initdb",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        initdb = handle.wait()

        if initdb:
            raise SubprocessError(
                "Airflow metadata database could not be initialized: `airflow initdb` failed",
                handle,
            )

        logging.debug("Completed `airflow initdb`")

    @hook("before_cleanup")
    def before_cleanup(self, invoker):
        config_file = invoker.files["config"]
        config_file.unlink()
        logging.debug(f"Deleted configuration at {config_file}")



import logging
import sys

import click
from meltano.core.db import project_engine
from meltano.core.error import SubprocessError
from meltano.core.plugin import PluginType
from meltano.core.plugin_invoker import invoker_factory
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.tracking import GoogleAnalyticsTracker

from . import cli
from .params import pass_project
from .utils import CliError

logger = logging.getLogger(__name__)


@cli.command(
    context_settings=dict(ignore_unknown_options=True, allow_interspersed_args=False)
)
@click.option(
    "--plugin-type", type=click.Choice(PluginType.cli_arguments()), default=None
)
@click.option(
    "--dump",
    type=click.Choice(["catalog", "config"]),
    help="Dump content of generated file",
)
@click.argument("plugin_name")
@click.argument("plugin_args", nargs=-1, type=click.UNPROCESSED)
@pass_project(migrate=True)
def invoke(project, plugin_type, dump, plugin_name, plugin_args):
    plugin_type = PluginType.from_cli_argument(plugin_type) if plugin_type else None

    _, Session = project_engine(project)
    session = Session()
    plugins_service = ProjectPluginsService(project)
    plugin = plugins_service.find_plugin(
        plugin_name, plugin_type=plugin_type, invokable=True
    )

    try:
        invoker = invoker_factory(project, plugin, plugins_service=plugins_service)
        with invoker.prepared(session):
            if dump:
                dump_file(invoker, dump)
                exit_code = 0
            else:
                handle = invoker.invoke(*plugin_args)
                exit_code = handle.wait()
    except SubprocessError as err:
        logger.error(err.stderr)
        raise
    finally:
        session.close()

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_invoke(
        plugin_name=plugin_name, plugin_args=" ".join(plugin_args)
    )

    sys.exit(exit_code)


def dump_file(invoker, file_id):
    try:
        content = invoker.dump(file_id)
        print(content)
    except FileNotFoundError as err:
        raise CliError(f"Could not find {file_id}") from err
    except Exception as err:
        raise CliError(f"Could not dump {file_id}: {err}") from err



import logging
import os
import re
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Dict, Iterable, List

from meltano.core.utils import NotFound
from meltano.core.utils import expand_env_vars as do_expand_env_vars
from meltano.core.utils import find_named, flatten

from .setting_definition import SettingDefinition, SettingMissingError
from .settings_store import SettingValueStore, StoreNotSupportedError

logger = logging.getLogger(__name__)


# sentinel value to use to prevent leaking sensitive data
REDACTED_VALUE = "(redacted)"


class SettingsService(ABC):
    LOGGING = False

    def __init__(self, project, show_hidden=True, env_override={}, config_override={}):
        self.project = project

        self.show_hidden = show_hidden

        self.env_override = env_override
        self.config_override = config_override

        self._setting_defs = None

    @property
    @abstractmethod
    def label(self):
        pass

    @property
    @abstractmethod
    def docs_url(self):
        pass

    @property
    def env_prefixes(self) -> [str]:
        """Return prefixes for setting environment variables."""
        return []

    @property
    @abstractmethod
    def db_namespace(self) -> str:
        """Return namespace for setting value records in system database."""
        pass

    @property
    @abstractmethod
    def setting_definitions(self) -> List[SettingDefinition]:
        """Return definitions of supported settings."""
        pass

    @property
    def inherited_settings_service(self):
        """Return settings service to inherit configuration from."""
        pass

    @property
    @abstractmethod
    def meltano_yml_config(self) -> Dict:
        """Return current configuration in `meltano.yml`."""
        pass

    @abstractmethod
    def update_meltano_yml_config(self, config):
        """Update configuration in `meltano.yml`."""
        pass

    @abstractmethod
    def process_config(self):
        """Process configuration dictionary to be used downstream."""
        pass

    @property
    def flat_meltano_yml_config(self):
        return flatten(self.meltano_yml_config, "dot")

    @property
    def env(self):
        return {**os.environ, **self.env_override}

    @classmethod
    def unredact(cls, values: dict) -> Dict:
        """
        Removes any redacted values in a dictionary.
        """

        return {k: v for k, v in values.items() if v != REDACTED_VALUE}

    def config_with_metadata(
        self,
        prefix=None,
        extras=None,
        source=SettingValueStore.AUTO,
        source_manager=None,
        **kwargs,
    ):
        if source_manager:
            source_manager.bulk = True
        else:
            source_manager = source.manager(self, bulk=True, **kwargs)

        config = {}
        for setting_def in self.definitions(extras=extras):
            if prefix and not setting_def.name.startswith(prefix):
                continue

            value, metadata = self.get_with_metadata(
                setting_def.name,
                setting_def=setting_def,
                source=source,
                source_manager=source_manager,
                **kwargs,
            )

            name = setting_def.name
            if prefix:
                name = name[len(prefix) :]

            config[name] = {**metadata, "value": value}

        return config

    def as_dict(self, *args, process=False, **kwargs) -> Dict:
        config_metadata = self.config_with_metadata(*args, **kwargs)

        if process:
            config = {
                key: metadata["setting"].post_process_value(metadata["value"])
                for key, metadata in config_metadata.items()
            }
            config = self.process_config(config)
        else:
            config = {
                key: metadata["value"] for key, metadata in config_metadata.items()
            }

        return config

    def as_env(self, *args, **kwargs) -> Dict[str, str]:
        full_config = self.config_with_metadata(*args, **kwargs)

        env = {}
        for key, config in full_config.items():
            value = config["value"]
            if value is None:
                continue

            setting_def = config["setting"]
            value = setting_def.stringify_value(value)

            for env_var in self.setting_env_vars(setting_def, for_writing=True):
                if env_var.negated:
                    continue

                env[env_var.key] = value

        return env

    def get_with_metadata(
        self,
        name: str,
        redacted=False,
        source=SettingValueStore.AUTO,
        source_manager=None,
        setting_def=None,
        expand_env_vars=True,
        **kwargs,
    ):
        try:
            setting_def = setting_def or self.find_setting(name)
        except SettingMissingError:
            pass

        if setting_def:
            name = setting_def.name

        self.log(f"Getting setting '{name}'")

        metadata = {"name": name, "source": source, "setting": setting_def}

        expandable_env = {**self.project.dotenv_env, **self.env}
        if setting_def and setting_def.is_extra:
            expandable_env.update(
                self.as_env(
                    extras=False,
                    redacted=redacted,
                    source=source,
                    source_manager=source_manager,
                )
            )

        manager = source_manager or source.manager(self, **kwargs)
        value, get_metadata = manager.get(name, setting_def=setting_def)
        metadata.update(get_metadata)

        if expand_env_vars and metadata.get("expandable", False):
            metadata["expandable"] = False

            expanded_value = do_expand_env_vars(value, env=expandable_env)

            if expanded_value != value:
                metadata["expanded"] = True
                metadata["unexpanded_value"] = value
                value = expanded_value

        if setting_def:
            if (
                setting_def.kind == "object"
                and metadata["source"] is SettingValueStore.DEFAULT
            ):
                object_value = {}
                object_source = SettingValueStore.DEFAULT
                for setting_key in [setting_def.name, *setting_def.aliases]:
                    flat_config_metadata = self.config_with_metadata(
                        prefix=f"{setting_key}.",
                        redacted=redacted,
                        source=source,
                        source_manager=source_manager,
                        expand_env_vars=expand_env_vars,
                    )
                    for nested_key, config_metadata in flat_config_metadata.items():
                        if nested_key in object_value:
                            continue

                        object_value[nested_key] = config_metadata["value"]

                        nested_source = config_metadata["source"]
                        if nested_source.overrides(object_source):
                            object_source = nested_source

                if object_value:
                    value = object_value
                    metadata["source"] = object_source

            cast_value = setting_def.cast_value(value)
            if cast_value != value:
                metadata["uncast_value"] = value
                value = cast_value

            # we don't want to leak secure informations
            # so we redact all `passwords`
            if redacted and value and setting_def.is_redacted:
                metadata["redacted"] = True
                value = REDACTED_VALUE

        self.log(f"Got setting '{name}' with metadata: {metadata}")
        return value, metadata

    def get_with_source(self, *args, **kwargs):
        value, metadata = self.get_with_metadata(*args, **kwargs)
        return value, metadata["source"]

    def get(self, *args, **kwargs):
        value, _ = self.get_with_source(*args, **kwargs)
        return value

    def set_with_metadata(
        self, path: List[str], value, store=SettingValueStore.AUTO, **kwargs
    ):
        self.log(f"Setting setting '{path}'")

        if isinstance(path, str):
            path = [path]

        name = ".".join(path)

        try:
            setting_def = self.find_setting(name)
        except SettingMissingError:
            setting_def = None

        metadata = {"name": name, "path": path, "store": store, "setting": setting_def}

        if value == REDACTED_VALUE:
            metadata["redacted"] = True
            return None, metadata

        if setting_def:
            cast_value = setting_def.cast_value(value)
            if cast_value != value:
                metadata["uncast_value"] = value
                value = cast_value

        manager = store.manager(self, **kwargs)
        set_metadata = manager.set(name, path, value, setting_def=setting_def)
        metadata.update(set_metadata)

        self.log(f"Set setting '{name}' with metadata: {metadata}")
        return value, metadata

    def set(self, *args, **kwargs):
        value, _ = self.set_with_metadata(*args, **kwargs)
        return value

    def unset(self, path: List[str], store=SettingValueStore.AUTO, **kwargs):
        self.log(f"Unsetting setting '{path}'")

        if isinstance(path, str):
            path = [path]

        name = ".".join(path)

        try:
            setting_def = self.find_setting(name)
        except SettingMissingError:
            setting_def = None

        metadata = {"name": name, "path": path, "store": store, "setting": setting_def}

        manager = store.manager(self, **kwargs)
        unset_metadata = manager.unset(name, path, setting_def=setting_def)
        metadata.update(unset_metadata)

        self.log(f"Unset setting '{name}' with metadata: {metadata}")
        return metadata

    def reset(self, store=SettingValueStore.AUTO, **kwargs):
        metadata = {"store": store}

        manager = store.manager(self, **kwargs)
        reset_metadata = manager.reset()
        metadata.update(reset_metadata)

        self.log(f"Reset settings with metadata: {metadata}")
        return metadata

    def definitions(self, extras=None) -> Iterable[Dict]:
        if self._setting_defs is None:
            self._setting_defs = [
                setting
                for setting in self.setting_definitions
                if setting.kind != "hidden" or self.show_hidden
            ]

        if extras is not None:
            return [
                setting
                for setting in self._setting_defs
                if (extras is True and setting.is_extra)  # noqa: WPS408
                or (extras is False and not setting.is_extra)
            ]

        return self._setting_defs

    def find_setting(self, name: str) -> SettingDefinition:
        try:
            return next(
                s for s in self.definitions() if s.name == name or name in s.aliases
            )
        except StopIteration as err:
            raise SettingMissingError(name) from err

    def setting_env_vars(self, setting_def, for_writing=False):
        return setting_def.env_vars(self.env_prefixes)

    def setting_env(self, setting_def):
        return self.setting_env_vars(setting_def)[0].key

    def log(self, message):
        if self.LOGGING:
            logger.debug(message)



import logging
import os
import secrets

import meltano.oauth.config
from flask import Flask, render_template, url_for
from meltano.core.project import Project
from werkzeug.contrib.fixers import ProxyFix


def create_app():
    project = Project.find()

    app = Flask(
        __name__, instance_path=str(project.root), instance_relative_config=True
    )

    app.config.from_object("meltano.oauth.config")
    app.config.from_mapping(**meltano.oauth.config.ProjectSettings(project).as_dict())

    @app.errorhandler(Exception)
    def _handle(e):
        logging.exception("An error occurred")
        return error()

    @app.route("/")
    def root():
        return render_template(
            "home.html",
            providers=(
                {
                    "label": "Facebook",
                    "url": url_for("OAuth.Facebook.login"),
                    "logo": url_for("static", filename="logos/facebook-logo.png"),
                },
                {
                    "label": "Google Ads",
                    "url": url_for("OAuth.GoogleAdwords.login"),
                    "logo": url_for("static", filename="logos/adwords-logo.png"),
                },
            ),
        )

    @app.route("/sample")
    def sample():
        return render_template("token.html", token=secrets.token_hex(128))

    @app.route("/error")
    def error():
        return render_template("error.html")

    from .providers import OAuth, facebook, google_adwords

    facebook(app)
    google_adwords(app)
    OAuth.init_app(app)

    # enable the parsing of X-Forwarded-* headers
    app.wsgi_app = ProxyFix(app.wsgi_app)

    return app



import logging
import os
from contextlib import contextmanager

import meltano.core.bundle as bundle
import yaml

from .project import Project
from .setting_definition import SettingDefinition

logger = logging.getLogger(__name__)


class ConfigService:
    def __init__(self, project: Project, use_cache=True):
        self.project = project

        self._settings = None
        self._current_meltano_yml = None
        self._use_cache = use_cache

    @property
    def settings(self):
        if self._settings is None:
            with bundle.find("settings.yml").open() as settings_yaml:
                settings = yaml.safe_load(settings_yaml)
            self._settings = list(map(SettingDefinition.parse, settings["settings"]))

        return self._settings

    @property
    def current_meltano_yml(self):
        if self._current_meltano_yml is None or not self._use_cache:
            self._current_meltano_yml = self.project.meltano
        return self._current_meltano_yml

    @contextmanager
    def update_meltano_yml(self):
        with self.project.meltano_update() as meltano_yml:
            yield meltano_yml

        self._current_meltano_yml = None

    @property
    def current_config(self):
        return self.current_meltano_yml.extras

    def update_config(self, config):
        with self.update_meltano_yml() as meltano_yml:
            meltano_yml.extras = config

    def make_meltano_secret_dir(self):
        os.makedirs(self.project.meltano_dir(), exist_ok=True)



import json
import logging
import shutil
import subprocess
from hashlib import sha1
from pathlib import Path
from typing import Dict

from jsonschema import Draft4Validator
from meltano.core.behavior.hookable import hook
from meltano.core.job import JobFinder, Payload
from meltano.core.plugin.error import PluginExecutionError, PluginLacksCapabilityError
from meltano.core.plugin_invoker import InvokerError
from meltano.core.setting_definition import SettingDefinition
from meltano.core.utils import file_has_data, flatten, merge, truthy

from . import PluginType, SingerPlugin
from .catalog import (
    MetadataExecutor,
    MetadataRule,
    SchemaExecutor,
    SchemaRule,
    property_breadcrumb,
    select_filter_metadata_rules,
    select_metadata_rules,
)

logger = logging.getLogger(__name__)


def config_metadata_rules(config):
    flat_config = flatten(config, "dot")

    rules = []
    for key, value in flat_config.items():
        # <tap_stream_id>.<key>
        # <tap_stream_id>.<prop>.<key>
        # <tap_stream_id>.<prop>.<subprop>.<key>
        # <tap_stream_id>.properties.<prop>.<key>
        # <tap_stream_id>.properties.<prop>.properties.<subprop>.<key>
        tap_stream_id, *props, key = key.split(".")

        rules.append(
            MetadataRule(
                tap_stream_id=tap_stream_id,
                breadcrumb=property_breadcrumb(props),
                key=key,
                value=value,
            )
        )

    return rules


def config_schema_rules(config):
    return [
        SchemaRule(
            tap_stream_id=tap_stream_id,
            breadcrumb=["properties", prop],
            payload=payload,
        )
        for tap_stream_id, stream_config in config.items()
        for prop, payload in stream_config.items()
    ]


class SingerTap(SingerPlugin):
    __plugin_type__ = PluginType.EXTRACTORS

    EXTRA_SETTINGS = [
        SettingDefinition(name="_catalog"),
        SettingDefinition(name="_state"),
        SettingDefinition(name="_load_schema", value="$MELTANO_EXTRACTOR_NAMESPACE"),
        SettingDefinition(name="_select", kind="array", value=["*.*"]),
        SettingDefinition(
            name="_metadata",
            aliases=["metadata"],
            kind="object",
            value={},
            value_processor="nest_object",
        ),
        SettingDefinition(
            name="_schema", kind="object", value={}, value_processor="nest_object"
        ),
        SettingDefinition(name="_select_filter", kind="array", value=[]),
    ]

    def exec_args(self, plugin_invoker):
        """
        Return the arguments list with the complete runtime paths.
        """
        args = ["--config", plugin_invoker.files["config"]]

        catalog_path = plugin_invoker.files["catalog"]
        if file_has_data(catalog_path):
            if "catalog" in plugin_invoker.capabilities:
                args += ["--catalog", catalog_path]
            elif "properties" in plugin_invoker.capabilities:
                args += ["--properties", catalog_path]
            else:
                logger.warn(
                    "A catalog file was found, but it will be ignored as the extractor does not advertise the `catalog` or `properties` capability"
                )

        state_path = plugin_invoker.files["state"]
        if file_has_data(state_path):
            if "state" in plugin_invoker.capabilities:
                args += ["--state", state_path]
            else:
                logger.warn(
                    "A state file was found, but it will be ignored as the extractor does not advertise the `state` capability"
                )

        return args

    @property
    def config_files(self):
        return {
            "config": "tap.config.json",
            "catalog": "tap.properties.json",
            "catalog_cache_key": "tap.properties.cache_key",
            "state": "state.json",
        }

    @property
    def output_files(self):
        return {"output": "tap.out"}

    @hook("before_invoke")
    def look_up_state_hook(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args or "--help" in exec_args:
            return

        try:
            self.look_up_state(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            pass

    def look_up_state(self, plugin_invoker, exec_args=[]):
        if "state" not in plugin_invoker.capabilities:
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support incremental state"
            )

        state_path = plugin_invoker.files["state"]

        try:
            # Delete state left over from different pipeline run for same extractor
            state_path.unlink()
        except FileNotFoundError:
            pass

        elt_context = plugin_invoker.context
        if not elt_context or not elt_context.job:
            # Running outside pipeline context: incremental state could not be loaded
            return

        if elt_context.full_refresh:
            logger.info(
                "Performing full refresh, ignoring state left behind by any previous runs."
            )
            return

        custom_state_filename = plugin_invoker.plugin_config_extras["_state"]
        if custom_state_filename:
            custom_state_path = plugin_invoker.project.root.joinpath(
                custom_state_filename
            )

            try:
                shutil.copy(custom_state_path, state_path)
                logger.info(f"Found state in {custom_state_filename}")
            except FileNotFoundError as err:
                raise PluginExecutionError(
                    f"Could not find state file {custom_state_path}"
                ) from err

            return

        # the `state.json` is stored in the database
        state = {}
        incomplete_since = None
        finder = JobFinder(elt_context.job.job_id)

        state_job = finder.latest_with_payload(elt_context.session, flags=Payload.STATE)
        if state_job:
            logger.info(f"Found state from {state_job.started_at}.")
            incomplete_since = state_job.ended_at
            if "singer_state" in state_job.payload:
                merge(state_job.payload["singer_state"], state)

        incomplete_state_jobs = finder.with_payload(
            elt_context.session, flags=Payload.INCOMPLETE_STATE, since=incomplete_since
        )
        for state_job in incomplete_state_jobs:
            logger.info(
                f"Found and merged incomplete state from {state_job.started_at}."
            )
            if "singer_state" in state_job.payload:
                merge(state_job.payload["singer_state"], state)

        if state:
            with state_path.open("w") as state_file:
                json.dump(state, state_file, indent=2)
        else:
            logger.warning("No state was found, complete import.")

    @hook("before_invoke")
    def discover_catalog_hook(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args or "--help" in exec_args:
            return

        try:
            self.discover_catalog(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            pass

    def discover_catalog(self, plugin_invoker, exec_args=[]):
        catalog_path = plugin_invoker.files["catalog"]
        catalog_cache_key_path = plugin_invoker.files["catalog_cache_key"]

        if catalog_path.exists():
            try:
                cached_key = catalog_cache_key_path.read_text()
                new_cache_key = self.catalog_cache_key(plugin_invoker)

                if cached_key == new_cache_key:
                    logger.debug(f"Using cached catalog file")
                    return
            except FileNotFoundError:
                pass

            logging.debug("Cached catalog is outdated, running discovery...")

        # We're gonna generate a new catalog, so delete the cache key.
        try:
            catalog_cache_key_path.unlink()
        except FileNotFoundError:
            pass

        custom_catalog_filename = plugin_invoker.plugin_config_extras["_catalog"]
        if custom_catalog_filename:
            custom_catalog_path = plugin_invoker.project.root.joinpath(
                custom_catalog_filename
            )

            try:
                shutil.copy(custom_catalog_path, catalog_path)
                logger.info(f"Found catalog in {custom_catalog_path}")
            except FileNotFoundError as err:
                raise PluginExecutionError(
                    f"Could not find catalog file {custom_catalog_path}"
                ) from err
        else:
            self.run_discovery(plugin_invoker, catalog_path)

        # test for the result to be a valid catalog
        try:
            with catalog_path.open("r") as catalog_file:
                catalog = json.load(catalog_file)
                schema_valid = Draft4Validator.check_schema(catalog)
        except Exception as err:
            catalog_path.unlink()
            raise PluginExecutionError(
                f"Catalog discovery failed: invalid catalog: {err}"
            ) from err

    def run_discovery(self, plugin_invoker, catalog_path):
        if not "discover" in plugin_invoker.capabilities:
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support catalog discovery (the `discover` capability is not advertised)"
            )

        try:
            with catalog_path.open("w") as catalog:
                result = plugin_invoker.invoke(
                    "--discover",
                    stdout=catalog,
                    stderr=subprocess.PIPE,
                    universal_newlines=True,
                )
                stdout, stderr = result.communicate()
                exit_code = result.returncode
        except Exception:
            catalog_path.unlink()
            raise

        if exit_code != 0:
            catalog_path.unlink()
            raise PluginExecutionError(
                f"Catalog discovery failed: command {plugin_invoker.exec_args('--discover')} returned {exit_code}: {stderr.rstrip()}"
            )

    @hook("before_invoke")
    def apply_catalog_rules_hook(self, plugin_invoker, exec_args=[]):
        if "--discover" in exec_args or "--help" in exec_args:
            return

        try:
            self.apply_catalog_rules(plugin_invoker, exec_args)
        except PluginLacksCapabilityError:
            pass

    def apply_catalog_rules(self, plugin_invoker, exec_args=[]):
        if (
            not "catalog" in plugin_invoker.capabilities
            and not "properties" in plugin_invoker.capabilities
        ):
            raise PluginLacksCapabilityError(
                f"Extractor '{self.name}' does not support entity selection or catalog metadata and schema rules"
            )

        config = plugin_invoker.plugin_config_extras

        schema_rules = []
        metadata_rules = []

        # If a custom catalog is provided, don't apply catalog rules
        if not config["_catalog"]:
            schema_rules.extend(config_schema_rules(config["_schema"]))

            metadata_rules.extend(select_metadata_rules(["!*.*"]))
            metadata_rules.extend(select_metadata_rules(config["_select"]))
            metadata_rules.extend(config_metadata_rules(config["_metadata"]))

        # Always apply select filters (`meltano elt` `--select` and `--exclude` options)
        metadata_rules.extend(select_filter_metadata_rules(config["_select_filter"]))

        if not schema_rules and not metadata_rules:
            return

        catalog_path = plugin_invoker.files["catalog"]
        catalog_cache_key_path = plugin_invoker.files["catalog_cache_key"]

        try:
            with catalog_path.open() as catalog_file:
                catalog = json.load(catalog_file)

            if schema_rules:
                SchemaExecutor(schema_rules).visit(catalog)

            if metadata_rules:
                MetadataExecutor(metadata_rules).visit(catalog)

            with catalog_path.open("w") as catalog_file:
                json.dump(catalog, catalog_file, indent=2)

            cache_key = self.catalog_cache_key(plugin_invoker)
            if cache_key:
                catalog_cache_key_path.write_text(cache_key)
            else:
                try:
                    catalog_cache_key_path.unlink()
                except FileNotFoundError:
                    pass
        except FileNotFoundError as err:
            raise PluginExecutionError(
                f"Applying catalog rules failed: catalog file is missing."
            ) from err
        except Exception as err:
            catalog_path.unlink()
            raise PluginExecutionError(
                f"Applying catalog rules failed: catalog file is invalid: {err}"
            ) from err

    def catalog_cache_key(self, plugin_invoker):
        # If the extractor is installed as editable, don't cache because
        # the result of discovery could change at any time.
        if plugin_invoker.plugin.pip_url.startswith("-e"):
            return None

        extras = plugin_invoker.plugin_config_extras

        # If a custom catalog is provided, there's no need to cache
        if extras["_catalog"]:
            return None

        # The catalog should be regenerated, and the catalog cache invalidated,
        # if any settings changed that could affect discovery, or if schema or
        # metadata rules changed.
        # Changes to selection rules and selection filter rules are ignored,
        # since "selected" metadata is reset using the `!*.*` selection rule anyway.
        key_dict = {
            **plugin_invoker.plugin_config,
            "_schema": extras["_schema"],
            "_metadata": extras["_metadata"],
        }

        key_json = json.dumps(key_dict)

        return sha1(key_json.encode()).hexdigest()



"""Defines SystemModel parent class for other models."""

from sqlalchemy import MetaData
from sqlalchemy.ext.declarative import declarative_base

SystemMetadata = MetaData()
SystemModel = declarative_base(metadata=SystemMetadata)



"""Add 'trigger' to Job

Revision ID: 367228df6a43
Revises: 87d9638f6ac6
Create Date: 2020-02-19 14:40:30.229756

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "367228df6a43"
down_revision = "87d9638f6ac6"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column("job", sa.Column("trigger", sa.String()))


def downgrade():
    op.drop_column("job", "trigger")



import functools
import os
import urllib
from pathlib import Path

import click
import click.globals
from meltano.core.db import project_engine
from meltano.core.migration_service import MigrationService
from meltano.core.project import Project
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.utils import pop_all

from .utils import CliError


def database_uri_option(func):
    @click.option("--database-uri", help="System database URI")
    def decorate(*args, database_uri=None, **kwargs):
        if database_uri:
            ProjectSettingsService.config_override["database_uri"] = database_uri

        return func(*args, **kwargs)

    return functools.update_wrapper(decorate, func)


class pass_project:  # noqa: N801
    """Pass current project to decorated CLI command function."""

    __name__ = "project"

    def __init__(self, migrate=False):
        self.migrate = migrate

    def __call__(self, func):
        @database_uri_option
        def decorate(*args, **kwargs):
            ctx = click.globals.get_current_context()

            project = ctx.obj["project"]
            if not project:
                raise CliError(
                    f"`{ctx.command_path}` must be run inside a Meltano project."
                    "\nUse `meltano init <project_name>` to create one."
                    "\nIf you are in a project, you may be in a subfolder. Navigate to the root directory."
                )

            # register the system database connection
            engine, _ = project_engine(project, default=True)

            if self.migrate:
                migration_service = MigrationService(engine)
                migration_service.upgrade(silent=True)
                migration_service.seed(project)

            func(project, *args, **kwargs)

        return functools.update_wrapper(decorate, func)



import json
import logging
import os

from meltano.core.behavior.hookable import hook
from meltano.core.db import project_engine
from meltano.core.plugin import BasePlugin
from meltano.core.project import Project
from meltano.core.utils import nest_object


class SingerPlugin(BasePlugin):
    def process_config(self, flat_config):
        non_null_config = {k: v for k, v in flat_config.items() if v is not None}
        return nest_object(non_null_config)

    @hook("before_configure")
    def before_configure(self, invoker, session):
        config_path = invoker.files["config"]
        with open(config_path, "w") as config_file:
            config = invoker.plugin_config_processed
            json.dump(config, config_file, indent=2)

        logging.debug(f"Created configuration at {config_path}")

    @hook("before_cleanup")
    def before_cleanup(self, invoker):
        config_path = invoker.files["config"]
        config_path.unlink()
        logging.debug(f"Deleted configuration at {config_path}")



from dotenv import dotenv_values
from meltano.core.settings_service import (
    SettingMissingError,
    SettingsService,
    SettingValueStore,
)
from meltano.core.utils import nest_object

from .config_service import ConfigService

UI_CFG_SETTINGS = {
    "ui.server_name": "SERVER_NAME",
    "ui.secret_key": "SECRET_KEY",
    "ui.password_salt": "SECURITY_PASSWORD_SALT",
}


class ProjectSettingsService(SettingsService):
    config_override = {}

    def __init__(self, *args, config_service: ConfigService = None, **kwargs):
        super().__init__(*args, **kwargs)

        self.config_service = config_service or ConfigService(self.project)

        self.env_override = {**self.project.env, **self.env_override}

        self.config_override = {
            **self.__class__.config_override,
            **self.config_override,
        }

    @property
    def label(self):
        return "Meltano"

    @property
    def docs_url(self):
        return "https://meltano.com/docs/settings.html"

    @property
    def env_prefixes(self):
        """Return prefixes for setting environment variables."""
        return ["meltano"]

    @property
    def db_namespace(self):
        """Return namespace for setting value records in system database."""
        return "meltano"

    @property
    def setting_definitions(self):
        """Return definitions of supported settings."""
        return self.config_service.settings

    @property
    def meltano_yml_config(self):
        """Return current configuration in `meltano.yml`."""
        return self.config_service.current_config

    def update_meltano_yml_config(self, config):
        """Update configuration in `meltano.yml`."""
        self.config_service.update_config(config)

    def process_config(self, config):
        """Process configuration dictionary for presentation in `meltano config meltano`."""
        return nest_object(config)

    def get_with_metadata(self, name: str, *args, **kwargs):
        value, metadata = super().get_with_metadata(name, *args, **kwargs)
        source = metadata["source"]

        if source is SettingValueStore.DEFAULT:
            # Support legacy `ui.cfg` files generated by `meltano ui setup`
            ui_cfg_value = self.get_from_ui_cfg(name)
            if ui_cfg_value is not None:
                value = ui_cfg_value
                metadata["source"] = SettingValueStore.ENV

        return value, metadata

    def get_from_ui_cfg(self, name: str):
        try:
            key = UI_CFG_SETTINGS[name]
            config = dotenv_values(self.project.root_dir("ui.cfg"))
            value = config[key]

            # Since `ui.cfg` is technically a Python file, `'None'` means `None`
            if value == "None":
                value = None

            return value
        except KeyError:
            return None



import copy
import fnmatch
import logging
import re
from collections import namedtuple
from enum import Enum
from typing import Dict, Iterable, Optional, Union

import yaml
from meltano.core.behavior import NameEq
from meltano.core.behavior.canonical import Canonical
from meltano.core.behavior.hookable import HookObject
from meltano.core.setting_definition import SettingDefinition
from meltano.core.utils import NotFound, compact, find_named, flatten

logger = logging.getLogger(__name__)


class VariantNotFoundError(Exception):
    def __init__(self, plugin: "PluginDefinition", variant_name):
        self.plugin = plugin
        self.variant_name = variant_name

    def __str__(self):
        return "{type} '{name}' variant '{variant}' is not known to Meltano. Variants: {variant_labels}".format(
            type=self.plugin.type.descriptor.capitalize(),
            name=self.plugin.name,
            variant=self.variant_name,
            variant_labels=self.plugin.variant_labels,
        )


class YAMLEnum(str, Enum):
    def __str__(self):
        return self.value

    @staticmethod
    def yaml_representer(dumper, obj):
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(obj))


yaml.add_multi_representer(YAMLEnum, YAMLEnum.yaml_representer)


class PluginType(YAMLEnum):
    EXTRACTORS = "extractors"
    LOADERS = "loaders"
    TRANSFORMS = "transforms"
    MODELS = "models"
    DASHBOARDS = "dashboards"
    ORCHESTRATORS = "orchestrators"
    TRANSFORMERS = "transformers"
    FILES = "files"

    def __str__(self):
        return self.value

    @property
    def descriptor(self):
        if self is self.__class__.FILES:
            return "file bundle"

        return self.singular

    @property
    def singular(self):
        """Makes it singular for `meltano add PLUGIN_TYPE`"""
        return self.value[:-1]

    @property
    def verb(self):
        if self is self.__class__.TRANSFORMS:
            return self.singular

        return self.value[:-3]

    @classmethod
    def value_exists(cls, value):
        return value in cls._value2member_map_

    @classmethod
    def cli_arguments(cls):
        args = [type.singular for type in cls]
        args.extend([type for type in cls])
        return args

    @classmethod
    def from_cli_argument(cls, value):
        if not value.endswith("s"):
            value += "s"

        return cls(value)


class PluginRef(Canonical):
    def __init__(self, plugin_type: Union[str, PluginType], name: str, **kwargs):
        self._type = (
            plugin_type
            if isinstance(plugin_type, PluginType)
            else PluginType(plugin_type)
        )

        super().__init__(name=name, **kwargs)

    @property
    def type(self):
        return self._type

    def __eq__(self, other):
        return self.name == other.name and self.type == other.type

    def __hash__(self):
        return hash((self.type, self.name))

    def set_presentation_attrs(self, extras):
        self.update(
            hidden=extras.pop("hidden", None),
            label=extras.pop("label", None),
            logo_url=extras.pop("logo_url", None),
            description=extras.pop("description", None),
        )


class Variant(NameEq, Canonical):
    ORIGINAL_NAME = "original"
    DEFAULT_NAME = "default"

    def __init__(
        self,
        name: str = None,
        original: Optional[bool] = None,
        deprecated: Optional[bool] = None,
        docs: Optional[str] = None,
        repo: Optional[str] = None,
        pip_url: Optional[str] = None,
        executable: Optional[str] = None,
        capabilities: Optional[list] = [],
        settings_group_validation: Optional[list] = [],
        settings: Optional[list] = [],
        **extras,
    ):
        super().__init__(
            name=name,
            original=original,
            deprecated=deprecated,
            docs=docs,
            repo=repo,
            pip_url=pip_url,
            executable=executable,
            capabilities=list(capabilities),
            settings_group_validation=list(settings_group_validation),
            settings=list(map(SettingDefinition.parse, settings)),
            extras=extras,
        )


class PluginDefinition(PluginRef):
    def __init__(
        self,
        plugin_type: PluginType,
        name: str,
        namespace: str,
        variant: Optional[str] = None,
        variants: Optional[list] = [],
        **extras,
    ):
        super().__init__(plugin_type, name)

        self._defaults["label"] = lambda p: p.name

        def default_logo_url(p):
            short_name = re.sub(r"^(tap|target)-", "", p.name)
            return f"/static/logos/{short_name}-logo.png"

        self._defaults["logo_url"] = default_logo_url

        if not variants:
            variant = Variant(variant, **extras)

            # Any properties considered "extra" by the variant should be
            # considered extras of the plugin definition.
            extras = variant.extras
            variant.extras = {}

            variants = [variant]

        # Attributes will be listed in meltano.yml in this order:
        self.namespace = namespace
        self.set_presentation_attrs(extras)
        self.extras = extras
        self.variants = list(map(Variant.parse, variants))

    def __iter__(self):
        for k, v in super().__iter__():
            if k == "variants" and len(v) == 1:
                # If there is only a single variant, its properties can be
                # nested in the plugin definition
                for variant_k, variant_v in v[0]:
                    if variant_k == "name":
                        variant_k = "variant"

                    yield (variant_k, variant_v)
            else:
                yield (k, v)

    def get_variant(self, variant_name: str) -> Variant:
        try:
            return find_named(self.variants, variant_name)
        except NotFound as err:
            raise VariantNotFoundError(self, variant_name) from err

    def find_variant(self, variant_or_name: Union[str, Variant] = None):
        if isinstance(variant_or_name, Variant):
            return variant_or_name

        if variant_or_name is None or variant_or_name == Variant.DEFAULT_NAME:
            return self.variants[0]

        if variant_or_name == Variant.ORIGINAL_NAME:
            try:
                return next(v for v in self.variants if v.original)
            except StopIteration:
                return self.variants[0]

        return self.get_variant(variant_or_name)

    def variant_label(self, variant):
        """Return label for specified variant."""
        variant = self.find_variant(variant)

        label = variant.name or Variant.ORIGINAL_NAME
        if variant == self.variants[0]:
            label = f"{label} (default)"
        elif variant.deprecated:
            label = f"{label} (deprecated)"

        return label

    @property
    def variant_labels(self):
        """Return labels for supported variants."""
        return ", ".join([self.variant_label(variant) for variant in self.variants])


class BasePlugin(HookObject):
    EXTRA_SETTINGS = []

    def __init__(self, plugin_def: PluginDefinition, variant: Variant):
        super().__init__()

        self._plugin_def = plugin_def
        self._variant = variant

    def __eq__(self, other):
        return (
            self._plugin_def == other._plugin_def  # noqa: WPS437
            and self._variant == other._variant  # noqa: WPS437
        )

    def __hash__(self):
        return hash((self._plugin_def, self._variant))

    def __iter__(self):
        yield from self._plugin_def

    def __getattr__(self, attr):
        try:
            return getattr(self._plugin_def, attr)
        except AttributeError:
            return getattr(self._variant, attr)

    @property
    def variant(self):
        return self._variant.name

    @property
    def executable(self):
        return self._variant.executable or self._plugin_def.name

    @property
    def extras(self):
        return {**self._plugin_def.extras, **self._variant.extras}

    @property
    def extra_settings(self):
        defaults = {f"_{k}": v for k, v in self.extras.items()}

        existing_settings = []
        for setting in self.EXTRA_SETTINGS:
            default_value = defaults.get(setting.name)
            if default_value is not None:
                setting = setting.with_attrs(value=default_value)

            existing_settings.append(setting)

        # Create setting definitions for unknown defaults,
        # including flattened keys of default nested object items
        existing_settings.extend(
            SettingDefinition.from_missing(
                existing_settings, defaults, custom=False, default=True
            )
        )

        return existing_settings

    def env_prefixes(self, for_writing=False):
        """Return environment variable prefixes to use for settings."""
        return [self.name, self.namespace]

    def is_installable(self):
        return self.pip_url is not None

    def is_invokable(self):
        return self.is_installable()

    def is_configurable(self):
        return True

    def should_add_to_file(self):
        return True

    @property
    def runner(self):
        return None

    def exec_args(self, files: Dict):
        return []

    @property
    def config_files(self):
        """Return a list of stubbed files created for this plugin."""
        return dict()

    @property
    def output_files(self):
        return dict()

    def process_config(self, config):
        return config



"""Defines helpers related to the system database."""

import logging
import time

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import text

from .project_settings_service import ProjectSettingsService

# Keep a Project → Engine mapping to serve
# the same engine for the same Project
_engines = dict()


def project_engine(project, default=False) -> ("Engine", sessionmaker):
    """Creates and register a SQLAlchemy engine for a Meltano project instance."""

    existing_engine = _engines.get(project)
    if existing_engine:
        return existing_engine

    settings = ProjectSettingsService(project)

    engine_uri = settings.get("database_uri")
    logging.debug(f"Creating engine {project}@{engine_uri}")
    engine = create_engine(engine_uri, pool_pre_ping=True)

    check_db_connection(
        engine,
        max_retries=settings.get("database_max_retries"),
        retry_timeout=settings.get("database_retry_timeout"),
    )

    init_hook(engine)

    engine_session = (engine, sessionmaker(bind=engine))

    if default:
        # register the default engine
        _engines[project] = engine_session

    return engine_session


def check_db_connection(engine, max_retries, retry_timeout):  # noqa: WPS231
    """Check if the database is available the first time a project's engine is created."""
    attempt = 0
    while True:
        try:
            engine.connect()
        except OperationalError:
            if attempt == max_retries:
                logging.error(
                    "Could not connect to the Database. Max retries exceeded."
                )
                raise
            attempt += 1
            logging.info(
                f"DB connection failed. Will retry after {retry_timeout}s. Attempt {attempt}/{max_retries}"
            )
            time.sleep(retry_timeout)
        else:
            break


def init_hook(engine):
    function_map = {"sqlite": init_sqlite_hook}

    try:
        function_map[engine.dialect.name](engine)
    except KeyError:
        pass
    except Exception as e:
        raise Exception(f"Can't initialize database: {str(e)}") from e


def init_sqlite_hook(engine):
    # enable the WAL
    engine.execute("PRAGMA journal_mode=WAL")


class DB:
    @classmethod
    def ensure_schema_exists(cls, engine, schema_name, grant_roles=()):
        """
        Make sure that the given schema_name exists in the database
        If not, create it

        :param db_conn: psycopg2 database connection
        :param schema_name: database schema
        """
        schema_identifier = schema_name
        group_identifiers = ",".join(grant_roles)

        create_schema = text(f"CREATE SCHEMA IF NOT EXISTS {schema_identifier}")
        grant_select_schema = text(
            f"ALTER DEFAULT PRIVILEGES IN SCHEMA {schema_identifier} GRANT SELECT ON TABLES TO {group_identifiers}"
        )
        grant_usage_schema = text(
            f"GRANT USAGE ON SCHEMA {schema_identifier} TO {group_identifiers}"
        )

        with engine.connect() as conn, conn.begin():
            conn.execute(create_schema)
            if grant_roles:
                conn.execute(grant_select_schema)
                conn.execute(grant_usage_schema)

        logging.info(f"Schema {schema_name} has been created successfully.")
        for role in grant_roles:
            logging.info(f"Usage has been granted for role: {role}.")
